# cython: language_level=3
# distutils: language = c++
from libcpp.list cimport list as clist
from libcpp.pair cimport pair
from libcpp.set cimport set as cset
from libcpp.stack cimport stack
from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.math cimport isnan

cdef extern from "extensions/FastFileIterator.h" nogil:
    cdef cppclass FastFileIterator[DATA_T]:
        FastFileIterator(const char*, size_t)
        DATA_T next()
        size_t size()
    int FastFileIteratorCompare[DATA_T](FastFileIterator[DATA_T]*,
                                        FastFileIterator[DATA_T]*)

# this is a least recently used cache written in C++ in an external file,
# exposing here so ManagedRaster can use it
cdef extern from "extensions/LRUCache.h" nogil:
    cdef cppclass LRUCache[KEY_T, VAL_T]:
        LRUCache(int)
        void put(KEY_T&, VAL_T&, clist[pair[KEY_T,VAL_T]]&)
        clist[pair[KEY_T,VAL_T]].iterator begin()
        clist[pair[KEY_T,VAL_T]].iterator end()
        bint exist(KEY_T &)
        VAL_T get(KEY_T &)

cdef extern from "extensions/ManagedRaster.h":
    cdef cppclass ManagedRaster:
        LRUCache[int, double*]* lru_cache
        cset[int] dirty_blocks
        int block_xsize
        int block_ysize
        int block_xmod
        int block_ymod
        int block_xbits
        int block_ybits
        long raster_x_size
        long raster_y_size
        int block_nx
        int block_ny
        int write_mode
        string raster_path
        int band_id
        int closed
        double nodata

        ManagedRaster() except +
        ManagedRaster(char*, int, bool) except +
        void set(long xi, long yi, double value)
        double get(long xi, long yi)
        void _load_block(int block_index) except *
        void close()

    cdef cppclass ManagedFlowDirRaster[T]:
        LRUCache[int, double*]* lru_cache
        cset[int] dirty_blocks
        int block_xsize
        int block_ysize
        int block_xmod
        int block_ymod
        int block_xbits
        int block_ybits
        long raster_x_size
        long raster_y_size
        int block_nx
        int block_ny
        int write_mode
        string raster_path
        int band_id
        int closed
        double nodata

        bint is_local_high_point(int xi, int yi)

        ManagedFlowDirRaster() except +
        ManagedFlowDirRaster(char*, int, bool) except +
        void set(long xi, long yi, double value)
        double get(long xi, long yi)
        void close()

    cdef cppclass D8
    cdef cppclass MFD

    cdef cppclass NeighborTuple:
        NeighborTuple() except +
        NeighborTuple(int, int, int, float) except +
        int direction, x, y
        float flow_proportion

    cdef cppclass DownslopeNeighborIterator[T]:
        ManagedFlowDirRaster[T] raster
        int col
        int row
        int n_dir
        int flow_dir
        int flow_dir_sum

        DownslopeNeighborIterator()
        DownslopeNeighborIterator(ManagedFlowDirRaster[T], int, int)
        void next()

    cdef cppclass DownslopeNeighborNoSkipIterator[T]:
        ManagedFlowDirRaster[T] raster
        int col
        int row
        int n_dir
        int flow_dir
        int flow_dir_sum

        DownslopeNeighborNoSkipIterator()
        DownslopeNeighborNoSkipIterator(ManagedFlowDirRaster[T], int, int)
        void next()

    cdef cppclass UpslopeNeighborIterator[T]:
        ManagedFlowDirRaster[T] raster
        int col
        int row
        int n_dir
        int flow_dir

        UpslopeNeighborIterator()
        UpslopeNeighborIterator(ManagedFlowDirRaster[T], int, int)
        void next()

    cdef cppclass UpslopeNeighborNoDivideIterator[T]:
        ManagedFlowDirRaster[T] raster
        int col
        int row
        int n_dir
        int flow_dir

        UpslopeNeighborNoDivideIterator()
        UpslopeNeighborNoDivideIterator(ManagedFlowDirRaster[T], int, int)
        void next()

    bint is_close(double, double)

    int[8] INFLOW_OFFSETS
    int[8] COL_OFFSETS
    int[8] ROW_OFFSETS
    int[8] FLOW_DIR_REVERSE_DIRECTION



import numpy
from osgeo import gdal

from ..geoprocessing import get_raster_info
from ..geoprocessing import raster_calculator
from ..geoprocessing_core import DEFAULT_GTIFF_CREATION_TUPLE_OPTIONS
from ..utils import gdal_use_exceptions


@gdal_use_exceptions
def extract_streams_d8(
        flow_accum_raster_path_band, flow_threshold, target_stream_raster_path,
        raster_driver_creation_tuple=DEFAULT_GTIFF_CREATION_TUPLE_OPTIONS):
    """Extract D8 streams based on a flow accumulation threshold.

    Creates an unsigned byte raster where pixel values of 1 indicate the
    presence of a stream and pixel values of 0 indicate the absence of a
    stream. Any flow accumulation pixels greater than ``flow_threshold`` are
    considered stream pixels. Nodata values found in the input flow
    accumulation raster propagate through to the target stream raster.

    Args:
        flow_accum_raster_path_band (tuple): A (path, band) tuple indicating
            the path to a D8 flow accumulation raster and the band index to
            use. Paths may use any GDAL-supported scheme, including virtual
            file system /vsi schemes.
        flow_threshold (number): The flow threshold. Flow accumulation values
            greater than this threshold are considered stream pixels, values
            less than this threshold are non-stream pixels.
        target_stream_raster_path (string): Where the target streams raster
            should be written.
        raster_driver_creation_tuple (tuple): A tuple where the first element
            is the GDAL driver name of the target raster and the second element
            is an iterable of raster creation options for the selected driver.

    Returns:
        ``None``
    """
    flow_accum_nodata = get_raster_info(
        flow_accum_raster_path_band[0])['nodata'][
            flow_accum_raster_path_band[1]-1]
    target_nodata = 255

    def _threshold_streams(flow_accum):
        out_matrix = numpy.full(flow_accum.shape, target_nodata,
                                dtype=numpy.uint8)

        valid_pixels = numpy.ones(flow_accum.shape, dtype=bool)
        if flow_accum_nodata is not None:
            valid_pixels = ~numpy.isclose(flow_accum, flow_accum_nodata)

        over_threshold = flow_accum > flow_threshold
        out_matrix[valid_pixels & over_threshold] = 1
        out_matrix[valid_pixels & ~over_threshold] = 0
        return out_matrix

    raster_calculator(
        [flow_accum_raster_path_band], _threshold_streams,
        target_stream_raster_path, gdal.GDT_Byte, target_nodata)
